import sys

import click

import sunspecdemo.clishared


@click.group(
    name='read',
    help='Read data points',
)
def group():
    pass


def common(device_factory, model_name, point_names):
    device = device_factory()
    model = device[model_name]

    if model is None:
        click.echo('No model found by the name {!r}'.format(model_name))

        click.echo(
            'Perhaps you meant one of: {}'.format(', '.join(device.models)),
        )

        sys.exit(1)

    model.read()

    show_all = len(point_names) == 0

    if show_all:
        points = {}
        points.update(model.model.points)
        points.update(model.model.points_sf)
    else:
        points = {
            point_name: model.model.points.get(
                point_name,
                model.model.points_sf.get(point_name),
            )
            for point_name in point_names
        }

    for point_name, point in points.items():
        click.echo('{}: {}'.format(
            point_name,
            point.value,
        ))

    if show_all:
        for index, block in enumerate(model.repeating):
            if block is None:
                continue

            click.echo()
            click.echo('Repetition: {}'.format(index))

            for point_name in block.points:
                point = block.block.points[point_name]
                click.echo('    {}: {}'.format(point_name, point.value))


commands = sunspecdemo.clishared.Commands.build(
    options=(
        sunspecdemo.clishared.model_name_option,
        sunspecdemo.clishared.point_names_option,
    ),
    common=common,
)
commands.add_to(group)



import time

import attr
import click
import sunspec.core.client

import sunspecdemo.clishared
import sunspecdemo.utils


def send_val(point, val):
    point.value_setter(val)
    point.write()


def clear_faults(config):
    model = config.state.block.model
    model.read_points()
    state_value = config.state.value_getter()
    print(f'Inverter State: {state_value}')
    #initialize flags to avoid clearing any flags that are already set
    config.cmd_flags.from_int(config.cmd_point.value_base)

    for attempt in range(10):
        # clear faults
        send_val(config.cmd_point, config.cmd_flags.set(config.fault_clear))
        time.sleep(0.5)
        model.read_points()
        state_value = config.state.value_getter()
        print(f'Inverter State: {state_value}')
        if state_value != config.faulted_value:
            break
    else:
        raise Exception('Unable to clear faults!')
    # remove fault clear command
    send_val(config.cmd_point, config.cmd_flags.clear(config.fault_clear))


def demo(device, config, cycles):
    # Read common model
    device.common.read()
    print(device.common)

    device.serial.read()
    print(device.serial)

    config.model.read()
    print(config.model)
    # Set control source, if defined
    if config.ctl_src is not None:
        send_val(config.ctl_src, 1)

    if config.invert_enable:
        value = config.dio_flags.set(config.invert_enable)
        send_val(config.dio_point, value)

    clear_faults(config=config)

    for ref in config.references:
        send_val(ref.point, ref.value)

    try:
        for _ in range(cycles):
            # enable and run
            if config.en_point: #enable point separate from other cmd flags
                send_val(config.en_point, 1)
            else:
                value = config.cmd_flags.set(config.enable)
                send_val(config.cmd_point, value)
                print('{}: {}'.format(value, config.cmd_flags.active()))

            config.model.read()
            print(config.model)

            time.sleep(0.5)
    finally:
        # remove run command
        if config.en_point: #enable point separate from other cmd flags
            send_val(config.en_point, 0)
        else:
            send_val(config.cmd_point, config.cmd_flags.clear(config.enable))
        # clear control source, if defined
        if config.ctl_src is not None:
            send_val(config.ctl_src, 0)

        print("controlset")


def gridtied_demo(device, invert_enable, cycles):
    points_basic = device.basic.model.points
    points_immediate = device.immediate.model.points
    refs = [
        Reference(point=points_basic['VRef'], value=480),
        Reference(point=points_basic['ECPNomHz'], value=60),
        Reference(point=points_immediate['WMaxLimPct'], value=25), #25% WMax
        Reference(point=points_immediate['VArMaxPct'], value=10), #10% VArMax
    ]
    config = DeviceConfig(
        en_point=points_immediate['Conn'],
        cmd_point=device.epc_control.model.points['CmdBits'],
        cmd_flags=sunspecdemo.utils.Flags(
            model=device.epc_control,
            point='CmdBits',
        ),
        dio_point=device.epc_control.model.points['DIO'],
        dio_flags=sunspecdemo.utils.Flags(
            model=device.epc_control,
            point='DIO',
        ),
        enable=None,
        fault_clear='FltClr',
        invert_enable='HwEnInv' if invert_enable else None,
        references=refs,
        ctl_src=device.epc_control.model.points['CtlSrc'],
        model=device.epc_control,
        state=device.inverter.model.points['StVnd'],
        faulted_value=3,
    )

    device.basic.read()
    print(device.basic)

    #enable real and reactive power commands
    send_val(points_immediate['WMaxLim_Ena'], 1)
    send_val(points_immediate['VArPct_Ena'], 1)
    #set var command mode to VArMax
    send_val(points_immediate['VArPct_Mod'], 2)

    config.cmd_flags.clear_all()

    demo(device=device, config=config, cycles=cycles)

    device.basic.read()
    print(device.basic)


def mayhem_demo(device, invert_enable, cycles):
    points = device.epc_control.model.points
    refs =  [
        Reference(point=points['CmdV'], value=480),
        Reference(point=points['CmdHz'], value=60),
        Reference(point=points['CmdRealPwr'], value=10000),
        Reference(point=points['CmdReactivePwr'], value=5000),
    ]
    config = DeviceConfig(
        en_point=None,
        cmd_point=device.epc_control.model.points['CmdBits'],
        cmd_flags=sunspecdemo.utils.Flags(
            model=device.epc_control,
            point='CmdBits',
        ),
        dio_point=device.epc_control.model.points['CmdBits'],
        dio_flags=sunspecdemo.utils.Flags(
            model=device.epc_control,
            point='CmdBits',
        ),
        enable='En',
        fault_clear='FltClr',
        invert_enable='InvertHwEnable' if invert_enable else None,
        references=refs,
        ctl_src=device.epc_control.model.points['CtlSrc'],
        model=device.epc_control,
        state=device.inverter.model.points['StVnd'],
        faulted_value=3,
    )

    demo(device=device, config=config, cycles=cycles)


def dcdc_demo(device, invert_enable, cycles):
    config = DeviceConfig(
        cmd_point=device.epc_control.model.points['CmdBits'],
        cmd_flags=sunspecdemo.utils.Flags(
            model=device.epc_control,
            point='CmdBits',
        ),
        dio_point=device.epc_control.model.points['DIO'],
        dio_flags=sunspecdemo.utils.Flags(
            model=device.epc_control,
            point='DIO',
        ),
        enable='En',
        fault_clear='FltClr',
        invert_enable='HwEnInv' if invert_enable else None,
        references=[
            References(
                point=device.epc_control.model.points['CmdVout'],
                value=800
            )
        ],
        ctl_src=device.epc_control.model.points['CtlSrc'],
        model=device.epc_control,
        state=device.epc_control.model.points['St'],
        faulted_value=3,
    )

    demo(device=device, config=config, cycles=cycles)


def abb_demo(device, invert_enable, cycles):
    points = device.abb_control.model.points
    refs =  [
        Reference(point=points['ABBCmdV'], value=480),
        Reference(point=points['ABBCmdHz'], value=60),
        Reference(point=points['ABBCmdRealPower'], value=10000), #10kW
        Reference(point=points['ABBCmdReactivePower'], value=5000), #5kVA
    ]
    config = DeviceConfig(
        en_point=None,
        cmd_point=device.abb_control.model.points['ABBCmdBits'],
        cmd_flags=sunspecdemo.utils.Flags(
            model=device.abb_control,
            point='ABBCmdBits',
        ),
        dio_point=device.epc_control.model.points['DIO'],
        dio_flags=sunspecdemo.utils.Flags(
            model=device.epc_control,
            point='DIO',
        ),
        enable='Enable',
        fault_clear='FaultReset',
        invert_enable=None,
        references=refs,
        ctl_src=None,
        model=device.abb_control,
        state=device.abb_control.model.points['ABBActiveState'],
        faulted_value=3, #probably wrong
    )

    send_val(points['ABBCmdRealPowerMax'], 20000)
    send_val(points['ABBCmdReactivePowerMax'], 10000)

    value = config.cmd_flags.clear_all()
    value = config.cmd_flags.set('GrdFlw_Pctrl')
    value = config.cmd_flags.set('GrdFlw_Qctrl')
    send_val(config.cmd_point, value)

    demo(device=device, config=config, cycles=cycles)


@click.group(
    help='Demo connection to a grid-tied converter',
)
@click.pass_obj
def gridtied(config):
    config.common = gridtied_demo


@click.group(
    help='Demo connection to a legacy grid-tied converter',
)
@click.pass_obj
def mh(config):
    config.common = mayhem_demo


@click.group(
    help='Demo connection to a DCDC converter',
)
@click.pass_obj
def dcdc(config):
    config.common = dcdc_demo


@click.group(
    help='Demo connection to an ABB converter',
)
@click.pass_obj
def abb(config):
    config.common = abb_demo


invert_enable_option = click.option(
    '--invert-enable/--no-invert-enable',
    help='Invert the converters hardware enable input',
)

cycles_option = click.option(
    '--cycles',
    type=int,
    default=25,
    help='Number of demo cycles to run',
)


@click.command(
    help='Demo direct serial Modbus RTU connection',
)
@sunspecdemo.clishared.serial_port_option
@sunspecdemo.clishared.serial_baudrate_option
@sunspecdemo.clishared.timeout_option
@sunspecdemo.clishared.model_path_option
@invert_enable_option
@sunspecdemo.clishared.slave_id_option
@cycles_option
@click.pass_obj
def serial(
        config,
        port,
        timeout,
        model_path,
        invert_enable,
        slave_id,
        baudrate,
        cycles,
):
    with sunspecdemo.utils.fresh_smdx_path(model_path):
        device = sunspec.core.client.SunSpecClientDevice(
            slave_id=slave_id,
            device_type=sunspec.core.client.RTU,
            name=port,
            baudrate=baudrate,
            timeout=timeout,
        )

    config.common(
        device=device,
        invert_enable=invert_enable,
        cycles=cycles,
    )


@click.command(
    help='Demo Modbus TCP connection',
)
@sunspecdemo.clishared.tcp_address_option
@sunspecdemo.clishared.tcp_port_option
@sunspecdemo.clishared.timeout_option
@sunspecdemo.clishared.model_path_option
@invert_enable_option
@sunspecdemo.clishared.slave_id_option
@cycles_option
@click.pass_obj
def tcp(
        config,
        address,
        port,
        timeout,
        model_path,
        invert_enable,
        slave_id,
        cycles,
):
    with sunspecdemo.utils.fresh_smdx_path(model_path):
        device = sunspec.core.client.SunSpecClientDevice(
            slave_id=slave_id,
            device_type=sunspec.core.client.TCP,
            ipaddr=address,
            ipport=port,
            timeout=timeout,
        )

    config.common(
        device=device,
        invert_enable=invert_enable,
        cycles=cycles,
    )


def add_commands(group):
    group.add_command(gridtied, name='gridtied')
    group.add_command(mh, name='mh')
    group.add_command(dcdc, name='dcdc')
    group.add_command(abb, name='abb')

    for group in (gridtied, mh, dcdc, abb):
        group.add_command(serial)
        group.add_command(tcp)


@attr.s
class Reference:
    point = attr.ib()
    value = attr.ib()


@attr.s
class DeviceConfig:
    en_point = attr.ib()
    cmd_point = attr.ib()
    cmd_flags = attr.ib()
    dio_point = attr.ib()
    dio_flags = attr.ib()
    enable = attr.ib()
    fault_clear = attr.ib()
    invert_enable = attr.ib()
    references = attr.ib()
    ctl_src = attr.ib()
    model = attr.ib()
    state = attr.ib()
    faulted_value = attr.ib()


import contextlib
import os

import attr
import sunspec.core.device


def apply_decorators(decorators):
    def decorator(f):
        for decorator in reversed(decorators):
            f = decorator(f)

        return f

    return decorator


def click_show_default_true():
    # https://github.com/pallets/click/issues/646#issuecomment-435317967

    import functools

    import click

    click.option = functools.partial(click.option, show_default=True)


@contextlib.contextmanager
def fresh_smdx_path(*paths):
    original_pathlist = sunspec.core.device.file_pathlist
    sunspec.core.device.file_pathlist = sunspec.core.util.PathList()

    for path in paths:
        sunspec.core.device.file_pathlist.add(os.fspath(path))

    try:
        yield sunspec.core.device.file_pathlist
    finally:
        sunspec.core.device.file_pathlist = original_pathlist


@attr.s
class Flags(object):
    _model = attr.ib()
    _point = attr.ib()
    _names = attr.ib(default=attr.Factory(set), converter=set)

    def __attrs_post_init__(self):
        self._symbols = self._model.model.points[self._point].point_type.symbols
        self._valid_names = set(s.id for s in self._symbols)

        self._validate_names(*self._names)

        self._bit_map = {int(s.value): s.id for s in self._symbols}

    def _validate_names(self, *names):
        if len(names) == 0:
            return

        names = set(names)

        bad_names = names - self._valid_names

        if len(bad_names) > 0:
            raise Exception(
                'Invalid flag{} specified for {}: {}'
                .format('s' if len(bad_names) > 1 else '',
                        self._point,
                        ', '.join(str(n) for n in bad_names)
                )
            )

    def set(self, *names):
        self._validate_names(*names)
        for name in names:
            self._names.add(name)
        return self.to_int()

    def clear(self, *names):
        self._validate_names(*names)
        for name in names:
            self._names.discard(name)
        return self.to_int()

    def to_int(self):
        return sum(1 << int(s.value) for s in self._symbols
                   if s.id in self._names)

    def set_all(self):
        self._names = set(self._valid_names)
        return self.to_int()

    def clear_all(self):
        self._names = set()
        return self.to_int()

    def from_int(self, source):
        s = '{:b}'.format(source)
        highest_bit = max(self._bit_map.keys())
        if len(s) > highest_bit+1:
            raise Exception(
                'Highest bit is {} but int has bit {} set'
                .format(highest_bit, len(s)-1)
            )

        names = (self._bit_map[n] for n, b in enumerate(reversed(s))
                 if b == '1')
        self.clear_all()
        self.set(*names)

    def active(self):
        return [s.id for s in self._symbols if s.id in self._names]



import socket

import attr
import click

import sunspec.core.client

import sunspecdemo.utils


@attr.s
class SerialDeviceFactory:
    port = attr.ib()
    baudrate = attr.ib()
    timeout = attr.ib()

    def build(self, slave_id):
        return sunspec.core.client.SunSpecClientDevice(
            slave_id=slave_id,
            device_type=sunspec.core.client.RTU,
            name=self.port,
            baudrate=self.baudrate,
            timeout=self.timeout,
        )


@attr.s
class TcpDeviceFactory:
    address = attr.ib()
    port = attr.ib()
    timeout = attr.ib()

    def build(self, slave_id):
        return sunspec.core.client.SunSpecClientDevice(
            slave_id=slave_id,
            device_type=sunspec.core.client.TCP,
            ipaddr=self.address,
            ipport=self.port,
            timeout=self.timeout,
        )


@click.group(
    help='Scan for responding nodes',
)
def cli():
    pass


node_id_range_option = click.option(
    '--slave-id-range',
    type=(int, int),
    default=(1, 246),
    help='Node ID range to scan for converters',
)


@cli.command(
    help='Scan over Modbus RTU',
)
@sunspecdemo.clishared.serial_port_option
@sunspecdemo.clishared.serial_baudrate_option
@sunspecdemo.clishared.timeout_option
@sunspecdemo.clishared.model_path_option
@node_id_range_option
def serial(
        port,
        timeout,
        model_path,
        slave_id_range,
        baudrate,
):
    device_factory = SerialDeviceFactory(
        port=port,
        baudrate=baudrate,
        timeout=timeout,
    )

    common(
        device_factory=device_factory,
        model_path=model_path,
        slave_id_range=slave_id_range,
    )


@cli.command(
    help='Scan over Modbus TCP',
)
@sunspecdemo.clishared.tcp_address_option
@sunspecdemo.clishared.tcp_port_option
@sunspecdemo.clishared.timeout_option
@sunspecdemo.clishared.model_path_option
@node_id_range_option
def tcp(
        address,
        port,
        timeout,
        model_path,
        slave_id_range,
):
    device_factory = TcpDeviceFactory(
        address=address,
        port=port,
        timeout=timeout,
    )

    common(
        device_factory=device_factory,
        model_path=model_path,
        slave_id_range=slave_id_range,
    )


timeout_exceptions = (
    sunspec.core.client.SunSpecClientError,
    socket.timeout,
)


def common(
        device_factory,
        model_path,
        slave_id_range,
):
    found = []

    for slave_id in range(slave_id_range[0], slave_id_range[1] + 1):
        click.echo('-- Scanning node {}'.format(slave_id))
        with sunspecdemo.utils.fresh_smdx_path(model_path):
            try:
                device_factory.build(
                    slave_id=slave_id,
                )
            except timeout_exceptions as e:
                click.echo('   ' + str(e))
                continue

        click.echo('   Found')
        found.append(slave_id)

    if len(found) > 0:
        found_message = '{nodes} found at {ids} {list}'.format(
            nodes='Nodes' if len(found) > 1 else 'Node',
            ids='IDs' if len(found) > 1 else 'ID',
            list=', '.join(str(id) for id in found),
        )
    else:
        found_message = 'No nodes found'

    click.echo()
    click.echo()
    click.echo('Scan complete.  {}'.format(found_message))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sunspecdemo']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3.0',
 'click>=7.0,<8.0',
 'importlib-metadata>=4.10.1,<5.0.0',
 'importlib>=1.0.4,<2.0.0',
 'pysunspec==2.0.0',
 'setuptools>=53.0.0',
 'toml>=0.10.2,<0.11.0',
 'toolz==0.9.0',
 'tqdm==4.32.1',
 'typing-extensions>=4.0.1,<5.0.0']

entry_points = \
{'console_scripts': ['sunspecdemo = sunspecdemo.cli:cli']}

setup_kwargs = {
    'name': 'sunspecdemo',
    'version': '0.1.95',
    'description': 'EPC SunSpec demonstration tool',
    'long_description': None,
    'author': 'Alex Anker',
    'author_email': 'alex.anker@epcpower.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '==3.7.9',
}


setup(**setup_kwargs)



import attr
import click
import serial.tools.list_ports

# https://github.com/pallets/click/issues/646#issuecomment-435317967
import sunspecdemo.utils
sunspecdemo.utils.click_show_default_true()

import sunspecdemo.demos
import sunspecdemo.models
import sunspecdemo.read
import sunspecdemo.scan
import sunspecdemo.write
import sunspecdemo.datalogger


@attr.s
class Config:
    common = attr.ib(default=None)


@click.group()
@click.version_option(version=sunspecdemo.__version__)
@click.pass_context
def cli(context):
    context.obj = Config()


sunspecdemo.demos.add_commands(group=cli)
cli.add_command(sunspecdemo.models.get_models)
cli.add_command(sunspecdemo.scan.cli, name='scan')
cli.add_command(sunspecdemo.read.group)
cli.add_command(sunspecdemo.write.group)
cli.add_command(sunspecdemo.datalogger.group)


@cli.command(
    name='list-ports',
    help='List available serial ports',
)
def list_ports():
    for port in serial.tools.list_ports.comports():
        print(port)
